/************************************************************************

Program to put ROM checksum in ROM image.

This program strips off the FreeBSD a.out header!

************************************************************************/
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

unsigned char *rom;
int romsize = 8192;

main(int argc, char **argv)
{
	int i, fd;
	unsigned int sum;
	char *progname;
	extern int optind;
	extern char *optarg;

	progname = argv[0];
	while ((i = getopt(argc, argv, "s:")) >= 0) {
		switch (i) {
		case 's':
			romsize = atoi(optarg);
			if (romsize <= 0)
				romsize = 8192;
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 1) {
		fprintf(stderr, "Usage: %s rom-file\n", progname);
		exit(1);
	}
	if ((fd = open(argv[0], O_RDWR)) < 0) {
		perror(argv[0]);
		exit(1);
	}
	if ((rom = malloc(romsize)) == 0) {
		fprintf(stderr, "Cannot malloc memory for ROM buffer\n");
		exit(1);
	}
	/* fill with FFs, slightly less work for PROM burner
	   and allows limited patching */
	memset(rom, 0xFF, romsize);
#ifndef __linux
	if (lseek(fd, (off_t)32, SEEK_SET) < 0) {
		perror("lseek");
		exit(1);
	}
#endif
	if (read(fd, rom, romsize) < 0) {
		perror("read");
		exit(1);
	}
	rom[5] = 0;
	for (i=0, sum=0; i < romsize; i++)
		sum += rom[i];
	rom[5] = -sum;
	for (i=0, sum=0; i < romsize; i++)
		sum += rom[i];
	if (sum & 0xFF)
		printf("checksum fails.\n");
	if (lseek(fd, (off_t)0, SEEK_SET) < 0) {
		perror("lseek");
		exit(1);
	}
	if (write(fd, rom, romsize) < 0) {
		perror(argv[0]);
		exit(1);
	}
	close(fd);
	exit(0);
}
