/* Stack needs to be end of bss + stacksize, but not exceeding lower 640kB */
#if RELOC+0xE000 > 0xA0000
#define STACKADDR	(0xA0000-RELOC)
#else
#define STACKADDR	0xE000
#endif
#define KERN_CODE_SEG	0x08
#define KERN_DATA_SEG	0x10
#define REAL_MODE_SEG	0x18
#define CR0_PE		1

#define opsize		.byte 0x66
#define addrsize	.byte 0x67

/* At entry, the processor is in 16 bit real mode and the code is being
 * executed from an address it was not linked to. Code must be pic and
 * 32 bit sensitive until things are fixed up.
 */
#ifdef __linux
_romstart:
#endif
#ifdef BOOTROM
	.word	0xaa55			/* bios extension signature */
	.byte	(ROMSIZE>>9)		/* no. of 512B blocks */
	jmp	1f			/* enter from bios here */
	.byte	0			/* checksum */
#ifdef	INT19H
1:	push	%eax
	push	%ds
	xor	%eax,%eax
	mov	%ax,%ds
	.byte	0xa1			/* MOV 0x304,%ax */
	.word	0x304
	.byte	0x3d			/* CMP $0x4d52, %ax  == 'MR' */
	.word	0x4d52
	jz	2f
	.byte	0xa1			/* MOV 0x64, %ax */
	.word	0x64
	.byte	0xa3			/* MOV %ax, 0x300 */
	.word	0x300
	.byte	0xa1			/* MOV 0x66, %ax */
	.word	0x66
	.byte	0xa3			/* MOV %ax, 0x302 */
	.word	0x302
	.byte	0xb8			/* MOV $_start-RELOCADDR, %ax */
#ifndef __linux
	.word	(_start-RELOC)
#else
	.word	(_start-_romstart)
#endif
	.byte	0xa3			/* MOV %ax, 0x64 */
	.word	0x64
	mov	%cs,%ax
	.byte	0xa3			/* MOV %ax, 0x66 */
	.word	0x66
	.byte	0xb8			/* MOV 'MR',%ax */
	.word	0x4d52
	.byte	0xa3			/* MOV %ax, 0x304 */
	.word	0x304
2:	pop	%ds
	pop	%eax
	lret
#else	/* take control now */
1:	mov	%ss,%dx			/* get current stack */
	mov	%sp,%ax
	mov	$(RELOC>>4),%bx		/* new ss */
	mov	$STACKADDR,%cx		/* new sp */
	mov	%bx,%ss
	mov	%cx,%sp
	push	%dx			/* save old stack on new stack */
	push	%ax
#endif	/* INT19H */
#endif	/* BOOTROM */

/**************************************************************************
START - Where all the fun begins....
**************************************************************************/
	.globl	_start
	.globl	start
start:
_start:
	cli
	cld
#ifdef BOOTROM				/* relocate ourselves */
        xor     %esi, %esi		/* zero for ROMs */
#else
	.byte	0xbe			/* MOV $0x100,%si -- 100h for .COM */
	.word	0x100
#endif
	xor	%edi,%edi
	.byte	0xb8			/* MOV $RELOCADDR>>4, %ax */
	.word	(RELOC>>4)
        mov     %ax, %es
	.byte	0xb9			/* MOV $ROMSIZE, %cx */
	.word   ROMSIZE
        cs
        rep
        movsb
	opsize
#ifndef __linux
	ljmp	$(RELOC>>4),$1f-RELOC		/* Jmp to RELOC:1f */
#else
	ljmp	$(RELOC>>4),$1f-_romstart	/* Jmp to RELOC:1f */
#endif
1:
	nop

	mov	%cs,%ax
	mov	%ax,%ds
	mov	%ax,%es
	mov	%ax,%ss
	.byte	0xb8			/* MOV $STACKADDR, %ax */
	.word	STACKADDR
	mov	%eax,%esp
	opsize
	call	_real_to_prot

#ifdef SERIAL_CONSOLE
	call	initserial
#endif
#ifdef	__ELF__
	call	main
#else
	call	_main
#endif
	.globl	_exit
	.globl	exit
exit:
_exit:
	call	_prot_to_real
#ifdef BOOTROM
#ifdef	INT19H
	xor	%eax,%eax
	mov	%ax,%ds
	.byte	0xa1			/* MOV 0x302, %ax */
	.word	0x302
	push	%eax
	.byte	0xa1			/* MOV 0x300, %ax */
	.word	0x300
	push	%eax
	lret
#else	/* INT19H */
/*	we reset sp to the location just before entering main
	instead of relying on the return from main because exit
	could have been called from anywhere */
	mov	$(STACKADDR-4),%cx
	mov	%cx,%sp
	pop	%ax			/* restore old stack */
	pop	%bx
	mov	%bx,%ss
	mov	%ax,%sp
	lret
#endif	/* INT19H */
#else
	int	$0x19
#endif

/**************************************************************************
CURRTICKS - Get Time
**************************************************************************/
	.globl	_currticks
	.globl	currticks
currticks:
_currticks:
/*	push	%ebp
	mov	%esp,%ebp */
	push	%ecx
	push	%edx
	xor	%edx,%edx
	call	_prot_to_real
	xor	%eax,%eax
	int	$0x1a
	opsize
	call	_real_to_prot
	xor	%eax,%eax
	shl	$16,%ecx
	mov	%edx,%eax
	or	%ecx,%eax
	pop	%edx
	pop	%ecx
/*	pop	%ebp */
	ret
#ifdef SERIAL_CONSOLE
/**************************************************************************
INITSERIAL - Initialize serial port. Stolen from lilo
**************************************************************************/
	.globl	_initserial
	.globl	initserial
sport:
	.byte	COMPORT	
sparam:
	.byte	COMPARM	/* 0xe3 == 9600N8, 0xa3 == 2400N8 */
slbase:	
	.word	0x0
initserial:
_initserial:
	push	%eax
	push	%ebx
	push	%ecx
	push	%edx
	xor	%ecx,%ecx
	movb	sport,%cl
	movb	sparam,%ch
	call	_prot_to_real
	opsize
	xor	%ax,%ax
	xor	%dx,%dx
	movb	%cl,%dl
	xchg	%ch,%al
	int	$0x14		/* initializes port. Base address of port is stored at 0x400 + offset */
	opsize
	call	_real_to_prot
	shl	%cx
	addl	$0x400,%ecx
	movw	(%ecx),%bx
	movw	%bx,slbase
	pop	%edx
	pop	%ecx
	pop	%ebx
	pop	%eax
	ret	
/**************************************************************************
SERDISP - Print a character on a serial line. Stolen from lilo.
**************************************************************************/
break:
	.byte	0
serdisp:
	push	%eax           /* wait for space in the send buffer */
	addw	$5,%dx
serwait:
	inb	%dx,%al
	testb	$0x10,%al	/* break -> set break flag */
	jz	nobrk
	movb	$1,break
nobrk:	
	testb	$0x20,%al	/* ready to send ? */
	jz	serwait		/* no -> wait */
	sub	$5,%dx		/* send the character */
	pop	%eax
	outb	%al,%dx
	ret
#endif	/* SERIAL_CONSOLE */
/**************************************************************************
PUTCHAR - Print a character
**************************************************************************/
	.globl	_putchar
	.globl	putchar
putchar:
_putchar:
	push	%ebp
	mov	%esp,%ebp
	push	%ecx
	push	%ebx
#ifdef SERIAL_CONSOLE
	push	%edx
	movw	slbase,%dx
	orw	%dx,%dx
	jz	sputchar_skip
	movb	8(%ebp),%eax
	call	serdisp
sputchar_skip:	
	pop	%edx
#endif
	movb	8(%ebp),%cl
	call	_prot_to_real
	opsize
	mov	$1,%ebx
	movb	$0x0e,%ah
	movb	%cl,%al
	int	$0x10
	opsize
	call	_real_to_prot
	pop	%ebx
	pop	%ecx
	pop	%ebp
	ret
/**************************************************************************
GETCHAR - Get a character
**************************************************************************/
	.globl	_getchar
	.globl	getchar
getchar:
_getchar:
	push	%ebx
#ifdef SERIAL_CONSOLE
	push	%edx
	movw	slbase,%dx
	orw	%dx,%dx
	jz	sgetchar_skip          /* no serial port available */
sgetchar_check_1:	
	addw	$5,%dx
sgetchar_check_2:	
	inb	%dx,%al
	testb	$1,%al
	jz	sgetchar_check_2
	subw	$5,%dx
	inb	%dx,%al
	andl	$0x7f,%eax
	je	sgetchar_check_1
	pop	%edx
	pop	%ebx
	ret
sgetchar_skip:	
	pop	%edx
#endif
	call	_prot_to_real
	movb	$0x0,%ah
	int	$0x16
	movb	%al,%bl
	opsize
	call	_real_to_prot
	xor	%eax,%eax
	movb	%bl,%al
	pop	%ebx
/*	pop	%ebp */
	ret

/**************************************************************************
ISKEY - Check for keyboard interrupt
**************************************************************************/
	.globl	_iskey
	.globl	iskey
iskey:
_iskey:
/*	push	%ebp
	mov	%esp,%ebp */
	push	%ebx
#ifdef SERIAL_CONSOLE
	push	%edx
	movw	slbase,%dx
	orw	%dx,%dx
	jz	siskey_skip	/* no serial port, we loose */
	addw	$5,%dx
	inb	%dx,%al
	andl	$0x01,%eax
	pop	%edx
	pop	%ebx
	ret
siskey_skip:	
	pop	%edx
#endif
	call	_prot_to_real
	xor	%ebx,%ebx
	movb	$0x1,%ah
	int	$0x16
	opsize
	jz	1f
	movb	%al,%bl
1:
	opsize
	call	_real_to_prot
	xor	%eax,%eax
	movb	%bl,%al
	pop	%ebx
/*	pop	%ebp */
	ret



/**************************************************************************
MEMSIZE - Determine size of extended memory
**************************************************************************/

	.globl	_memsize
	.globl	memsize
memsize:
_memsize:
	push	%ebx
	call	_prot_to_real
	.byte	0xb8
	.word	0x8800
	int	$0x15
	mov	%ax,%bx
	opsize
	call	_real_to_prot
	xor	%eax,%eax
	mov	%bx,%ax
	pop	%ebx
	ret

/**************************************************************************
START_LINUX - Call linux starter code (real mode)
**************************************************************************/
	.globl	_start_linux
	.globl	start_linux
start_linux:
_start_linux:
	call	_prot_to_real
	opsize
	call	1f
	opsize
	call	_real_to_prot
	ret
1:	opsize
	ljmp	$0x9020,$0x0

	.globl	_xstart
	.globl	xstart

xstart:  /* FIXME: this code is rather ugly! could some x86 programmer please*/
_xstart: /* come up with something better (I have no clue of x86 programming)*/
	pushl	%ebx
	pushl	%ecx
        pushl   %edx
	movw	16(%esp,1),%eax /* ugly self-modifying code! */
	movw	%eax,2f+2
	movw	$0x0,2f+4
	movw	18(%esp,1),%eax
	movw	%eax,2f+6
	movl	24(%esp,1),%ecx
	movl	20(%esp,1),%ebx
	call	_prot_to_real
	opsize
	ljmp	$(RELOC>>4),$1f-_romstart /* flush cache? */
1:	nop
	opsize
	pushl	%ecx   /* bootp record */
	opsize
	pushl	%ebx   /* file header */
        opsize
        movl    $((RELOC<<12)+(3f-_romstart)),%eax
        opsize
        pushl   %eax
2:      opsize
        ljmp    $(RELOC>>4),$3f-_romstart
3:      opsize
	popl	%eax
	opsize
	popl	%eax
	opsize	
	call	_real_to_prot
        popl    %edx
	popl	%ecx
	popl	%ebx
	ret
        	
/*
 * C library -- _setjmp, _longjmp
 *
 *	longjmp(a,v)
 * will generate a "return(v)" from the last call to
 *	setjmp(a)
 * by restoring registers from the stack.
 * The previous signal state is restored.
 */

	.globl	_setjmp
	.globl	setjmp
setjmp:
_setjmp:
	movl	4(%esp),%ecx 
	movl	0(%esp),%edx
	movl	%edx, 0(%ecx)
	movl	%ebx, 4(%ecx)
	movl	%esp, 8(%ecx)
	movl	%ebp,12(%ecx)
	movl	%esi,16(%ecx)
	movl	%edi,20(%ecx)
	movl	%eax,24(%ecx)
	movl	$0,%eax
	ret

	.globl	_longjmp
	.globl	longjmp
longjmp:
_longjmp:
	movl	4(%esp),%edx
	movl	8(%esp),%eax
	movl	0(%edx),%ecx
	movl	4(%edx),%ebx
	movl	8(%edx),%esp
	movl	12(%edx),%ebp
	movl	16(%edx),%esi
	movl	20(%edx),%edi
	cmpl	$0,%eax
	jne	1f
	movl	$1,%eax
1:	movl	%ecx,0(%esp)
	ret

/**************************************************************************
___MAIN - Dummy to keep GCC happy (do we still need it?)
**************************************************************************/
#if	0
	.globl	___main
	.globl	__main
__main:
___main:
	ret
#endif

/**************************************************************************
REAL_TO_PROT - Go from REAL mode to Protected Mode
**************************************************************************/
	.globl	_real_to_prot
	.globl	real_to_prot
real_to_prot:
_real_to_prot:
	cli
	cs
	addrsize
#ifndef __linux
	lgdt	gdtarg-RELOC 
#else
	lgdt	gdtarg-_romstart
#endif
	mov	%cr0, %eax
	opsize
	or	$CR0_PE, %eax
	mov	%eax, %cr0		/* turn on protected mode */

	/* jump to relocation, flush prefetch queue, and reload %cs */
	opsize
	ljmp	$KERN_CODE_SEG, $1f
1:
	/* reload other segment registers */
	movl	$KERN_DATA_SEG, %eax
	movl	%ax, %ds
	movl	%ax, %es
	movl	%ax, %ss
	add	$RELOC,%esp		/* Fix up stack pointer */
	pop	%eax			/* Fix up return Address */
	add	$RELOC,%eax
	push	%eax
	ret


/**************************************************************************
PROT_TO_REAL - Go from Protected Mode to REAL Mode
**************************************************************************/
	.globl	_prot_to_real
	.globl	prot_to_real
prot_to_real:
_prot_to_real:
	pop	%eax
	sub	$RELOC,%eax		/* Adjust return address */
	push	%eax
	sub	$RELOC,%esp		/* Adjust stack pointer */
	ljmp	$REAL_MODE_SEG, $1f	/* jump to a 16 bit segment */
1:
	/* clear the PE bit of CR0 */
	mov	%cr0, %eax
	opsize
	andl 	$0!CR0_PE, %eax
	mov	%eax, %cr0

	/* make intersegment jmp to flush the processor pipeline
	 * and reload CS register
	 */
	opsize
#ifndef __linux
	ljmp	$(RELOC)>>4, $2f-RELOC 
#else
	ljmp	$(RELOC)>>4, $2f-_romstart
#endif
2:
	/* we are in real mode now
	 * set up the real mode segment registers : DS, SS, ES
	 */
	mov	%cs, %ax
	mov	%ax, %ds
	mov	%ax, %es
	mov	%ax, %ss
	sti
	opsize
	ret

/**************************************************************************
GLOBAL DESCRIPTOR TABLE
**************************************************************************/
	.align	4
gdt:
	.word	0, 0
	.byte	0, 0x00, 0x00, 0

	/* code segment */
	.word	0xffff, 0
	.byte	0, 0x9f, 0xcf, 0

	/* data segment */
	.word	0xffff, 0
	.byte	0, 0x93, 0xcf, 0

	/* 16 bit real mode */
	.word	0xffff, 0
	.byte	0, 0x9b, 0x0f, 0

	.align	4
gdtarg:
	.word	0x1f			/* limit */
	.long	gdt			/* addr */
